package byu;

/****************************************************************************
** COPYRIGHT (C):    1996 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Computing Concepts with Java
** FILE:             GraphicsApplet.java
****************************************************************************/

/**
 * An applet for drawing simple graphics and getting user input
 * @version 1.00 11 Apr 1997
 * @author Cay Horstmann
 * @author Quinn Snell (BYU)
 */

import java.awt.event.*;

public abstract class GraphicsApplet extends java.applet.Applet implements Runnable, MouseListener, MouseMotionListener, KeyListener, ActionListener
{  /**
	* Constructs a graphics applet
	*/

   Thread runner;
   Point clicked;
   java.awt.TextField textbox;
   GraphicsCanvas graph;
   String text;
   java.awt.Label label;
   java.awt.Label titlebar;
   String tracking;
   GraphicsCoord coord;


   private static GraphicsApplet theApplet;
   public GraphicsApplet()
   {  theApplet = this;
	  coord = new GraphicsCoord();
   }   
   public synchronized void actionPerformed(java.awt.event.ActionEvent evt)
   {
	  String cmd = evt.getActionCommand();
	  //System.out.println("Action: " + cmd);
	  //System.out.flush();
	  if (cmd.equals("Enter"))
	  {  text = textbox.getText();
		 textbox.setText("");
		 notify();
	  }
   }   
   private void add(java.awt.Component c, java.awt.GridBagLayout gbl,
		 java.awt.GridBagConstraints gbc, int x, int y, int w, int h)
   {  
	  gbc.gridx = x;
	  gbc.gridy = y;
	  gbc.gridwidth = w;
	  gbc.gridheight = h;
	  gbl.setConstraints(c, gbc);
	  add(c);
   }   
   public static void addShape(Shape s) { theApplet.graph.addShape(s); }   
   /**
	* Clears the drawing window 
	*/

   public static void clearWindow()
   {  theApplet.graph.shapes = new java.util.Vector();
	  theApplet.repaint();
   }   
   public static void erase(Shape s) { theApplet.graph.erase(s); }   
   GraphicsCoord getCoord()
   {  coord.setPixel(graph.getSize().width, graph.getSize().height);
	  return coord;
   }   
   public void init()
   {  java.awt.GridBagLayout gbl = new java.awt.GridBagLayout();
	  setLayout(gbl);

	  java.awt.GridBagConstraints gbc = new java.awt.GridBagConstraints();
	  gbc.anchor = java.awt.GridBagConstraints.NORTHWEST;

	  gbc.weightx = 1000;
	  gbc.weighty = 100;
	  gbc.fill = java.awt.GridBagConstraints.BOTH;
	  graph = new GraphicsCanvas();
	  add(graph, gbl, gbc, 0, 0, 2, 1);

	  gbc.weighty = 0;
	  gbc.fill = java.awt.GridBagConstraints.HORIZONTAL;
	  label = new java.awt.Label();
	  add(label, gbl, gbc, 0, 1, 2, 1);

	  textbox = new java.awt.TextField();
	  add(textbox, gbl, gbc, 0, 2, 1, 1);

	  gbc.weightx = 0;
	  java.awt.Button enter = new java.awt.Button("Enter");
	  add(enter, gbl, gbc, 1, 2, 1, 1);

	  titlebar = new java.awt.Label("BYU Introductory Programming", 
											 java.awt.Label.CENTER);
	  add(titlebar, gbl, gbc, 0, 3, 2, 1);

	  graph.addMouseListener(this);
	  graph.addMouseMotionListener(this);
	  textbox.addKeyListener(this);
	  enter.addActionListener(this);
   }   
   //Other unused key methods
   public synchronized void keyPressed(KeyEvent evt) {;}   
   public synchronized void keyReleased(KeyEvent evt){;}   
   public synchronized void keyTyped(KeyEvent evt)
   { 
	  char key = evt.getKeyChar();
	  //System.out.println("Key: [" + key + "]");
	  //System.out.flush();
	  if (key == '\n')
	  {  text = textbox.getText();
		 textbox.setText("");
		 notify();
	  }
   }   
   public synchronized void mouseClicked(java.awt.event.MouseEvent evt)
   {
	  int x, y;
	  x = evt.getX();
	  y = evt.getY();
	  if (tracking != null)
	  {  java.awt.Rectangle bounds = graph.getBounds();
		 if (bounds.contains(x, y))
		 {  x -= graph.getBounds().x;
			y -= graph.getBounds().y;
			clicked = new Point(coord.xFromPixelPos(x),
			   coord.yFromPixelPos(y));
			tracking = null;
			notify();
		 }
	  }
   }   
   public synchronized void mouseDragged(MouseEvent e) {;}   
   public void mouseEntered(MouseEvent e) {;}   
   public void mouseExited(MouseEvent e) {;}   
   public synchronized void mouseMoved(MouseEvent evt)
   {
	  int x, y;
	  x = evt.getX();
	  y = evt.getY();
	  if (tracking != null)
	  {  java.awt.Rectangle bounds = graph.getBounds();
		 if (bounds.contains(x, y))
		 {  x -= bounds.x;
			y -= bounds.y;
			label.setText(tracking + " (" + coord.xFromPixelPos(x) + ","
			   + coord.yFromPixelPos(y) + ")");
		 }
		 else
			label.setText(tracking);
	  }
   }   
   //Other unused mouse methods
   public synchronized void mousePressed(MouseEvent e) {;}   
   public synchronized void mouseReleased(MouseEvent e){;}   
   /**
	* Prompts user for a floating point input
	* @param s the prompt string
	* @return the floating point number the user typed in response
	*/

   public static double readDouble(String s)
   {  try
	  {  return Double.valueOf(readString(s).trim()).doubleValue();
	  }
	  catch(NumberFormatException e)
	  {  return 0;
	  }
   }   
   private synchronized String readInput(String s)
   {  label.setText(s);
	  textbox.requestFocus();
	  while (text == null)
	  {  try { wait(); } catch(InterruptedException e) {}
	  }
	  String r = text;
	  text = null;
	  label.setText("");

	  return r;
   }   
   /**
	* Prompts user for an integer input
	* @param s the prompt string
	* @return the integer the user typed in response
	*/

   public static int readInt(String s)
   {  try
	  {  return Integer.valueOf(readString(s).trim()).intValue();
	  }
	  catch(NumberFormatException e)
	  {  return 0;
	  }
   }   
   /**
	* Prompts user for mouse input
	* @param s the prompt string
	* @return the position at which the user clicked the mouse in response
	*/

   public synchronized Point readMouse(String s)
   {  tracking = s;
	  label.setText(s);
	  while (clicked == null)
	  {  try { wait(); } catch(InterruptedException e) {}
	  }
	  Point r = clicked;
	  clicked = null;
	  label.setText("");
	  return r;
   }   
   /**
	* Prompts user for a string input
	* @param s the prompt string
	* @return the string the user typed in response
	*/

   public static String readString(String s)
   {  return theApplet.readInput(s);
   }   
   /**
	* Define this method in your graphics applet class to draw graphics and get user input
	*/

   public abstract void run();   
   /**
	* Sets the coordinate system for drawing graphics
	* @param xleft the x coordinate of the left border
	* @param ytop the y coordinate of the top border
	* @param xright the x coordinate of the right border
	* @param ybottom the y coordinate of the bottom border
	*/

   public void setCoord(double xleft, double ytop, double xright, double ybot)
   {  coord.setCoord(xleft, ytop, xright, ybot);
	  coord.setPixel(graph.getSize().width, graph.getSize().height);
   }   
   public void start()
   {  if (runner == null)
	  {  runner = new Thread(this);
		 runner.start();
	  }
	  else if (runner.isAlive())
		 runner.resume();
   }   
   public void stop()
   {  if (runner != null && runner.isAlive())
		 runner.suspend();
   }   
}
