package byu;

/****************************************************************************
** COPYRIGHT (C):    1996 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Computing Concepts with Java
** FILE:             Message.java
****************************************************************************/

/**
 * A messge for labeling simple graphics
 * @version 1.00 11 Apr 1997
 * @author Cay Horstmann
 */

public class Message extends Shape
{  /**
	* Creates a new message object
	* @param p the top left corner of the message
	* @param s the text of the message
	*/

   private Point start;
   private String text;
   public Message(Point p, String s)
   {  start = p;
	  text = s;
   }   
   /**
	* Creates a copy of this object
	* @return a copy of this object
	*/

   public Object clone()
   {  Message r = (Message)super.clone();
	  r.start = (Point)start.clone();
	  return r;
   }   
   public void draw()
   {
	  GraphicsApplet.addShape(this);
   }   
   /**
	* Tests if this object equals another
	* @param b the object to compare with
	* @return <tt>true</tt> if the two objects are equal
	*/

   public boolean equals(Object b)
   {  if (b instanceof Message)
	  {  Message pb = (Message)b;
		 return start.equals(pb.start) && text.equals(pb.text);
	  }
	  else return false;
   }   
   /**
	* Gets the top left corner of the message 
	* @return the top left corner of the message 
	*/

   public Point getStart()
   {  return start;
   }   
   /**
	* Gets the text of the message 
	* @return the text of the message 
	*/

   public String getText()
   {  return text;
   }   
   /**
	* Computes a hash code for this object
	* @return a hash code for this object
	*/

   public int hashCode()
   {  return 29 * start.hashCode() + 31 * text.hashCode();
   }   
   /**
	* Moves this graphical shape
	* @param dx the amount to move by in x direction
	* @param dy the amount to move by in y direction
	*/

   public void move(double x, double y)
   {  start.move(x, y);
   }   
   void redraw(java.awt.Graphics g, GraphicsCoord coord)
   {  int y = coord.yToPixelPos(start.getY());
	  y += g.getFontMetrics().getAscent();
		 g.drawString(text, coord.xToPixelPos(start.getX()), y);
   }   
   /**
	* Creates a string representation of this object
	* @return a string representation of this object
	*/

   public String toString()
   {  return "Message[start = " + start.toString() + "," + "text = " + text + "]";
   }   
}
