package EDU.auburn.VGJ;

import EDU.auburn.VGJ.algorithm.tree.TreeAlgorithm;
import EDU.auburn.VGJ.graph.*;
import EDU.auburn.VGJ.graph.Node;
import EDU.auburn.VGJ.gui.GraphWindow;

import clad.dna;

import java.applet.Applet;

import java.awt.*;
import java.awt.Button;
import java.awt.Event;
import java.awt.Font;
import java.awt.Frame;
import java.io.File;
import java.util.Vector;

/*
    File: VGJ.java
    Date      Author
    2/16/97   Larry Barowski
  */
import java.awt.GridLayout;
import java.awt.event.*;

import javax.swing.*;


/**
 *  The VGJ applet. It is a big button that pops up VGJ graph editor windows.
 *  </p>Here is the <a href="../VGJ.java">source</a> .
 *
 * @author     Larry Barowski
 * @created    June 3, 2003
 */
public class VGJ
		 extends Applet
{

	private int appCount_   = 0;
	private boolean isApp_  = false;



	/**
	 *  Description of the Method
	 *
	 * @param  sourceDirectory  Description of the Parameter
	 * @param  suffix           Description of the Parameter
	 * @return                  Description of the Return Value
	 */
	private static File[] filterFiles(File sourceDirectory, String suffix)
	{
		/*
		    Capture all sourceDirectory files matching the filter criteria.
		    lastModifedTime is global.
		  */
		//get a list of all files in the source directory
		String[] allFileNames  = sourceDirectory.list();

		//filter the files
		Vector filteredFiles   = new Vector();
		for (int i = 0; i < allFileNames.length; i++)
		{
			File file  = new File(sourceDirectory, allFileNames[i]);

			if (file.isFile() && file.canRead() && file.getName().endsWith(suffix))
			{
				filteredFiles.addElement(file);
			}
		}

		//put the filtered files in an array for more convenient access later
		File[] files           = new File[filteredFiles.size()];
		for (int i = 0; i < filteredFiles.size(); i++)
		{
			files[i] = (File) filteredFiles.elementAt(i);
		}

		return files;
	}


	/**
	 *  The main program for the VGJ class
	 *
	 * @param  args  The command line arguments
	 */
	public static void main(String[] args)
	{

		//   Frame aframe = new Frame();
		//ImageIcon openIcon = new ImageIcon("TCSicon.gif");

		/*
		    JOptionPane.showMessageDialog(aframe,"TCS is a program for the estimation of haplotype cladograms." +		    "\nWritten by Mark Clement and David Posada" +		    "\nDepartments of Computer Science (MC) and Zoology (DP), " +		    "\nBrigham Young University, Provo, Utah, USA." +		    "\n\nTCS uses the freeware VGJ (see docs)" +		    "\n\nPLEASE REALIZE THAT THIS IS AN ALPHA VERSION." +		    "\nTHIS MEANS THAT BUGS ARE LIKELY..." +		    "\nIT IS RECOMMEND THAT YOU CHECK THE RESULTS",		    "TCS " + VERSION ,  JOptionPane.INFORMATION_MESSAGE,openIcon);
		  */
		boolean saveStandard  = false;
		// String directoryName;
		if (args.length != 0)
		{

			boolean saveAsPict    = false;
			String fileToRun      = args[0];
			if (fileToRun.equals("-toPict"))
			{
				saveAsPict = true;
				fileToRun = args[1];

			}
			else
			{
				if (fileToRun.equals("-saveStand"))
				{
					saveStandard = true;
					fileToRun = args[1];

				}
			}

			File sourceDirectory  = new File(fileToRun);

			File[] files          = null;
			GraphWindow gw        = null;
			if (!saveAsPict)
			{
				// System.out.println("Will use phylip files");
				System.out.println("Will use phylip and graph files");
				files = filterFiles(sourceDirectory, ".phylip");
			}
			else
			{
				System.out.println("Will use graph files");
				files = filterFiles(sourceDirectory, ".graph");

			}
			gw = buildWindow_(true);
			gw.hide();

			// do this for each file...
			for (int x = 0; x < files.length; x++)
			{
				String file  = sourceDirectory.getAbsolutePath() + "/" + files[x].getName();
				System.out.println("TCS processing: " + file);
				if (saveAsPict)
				{
					gw.PictSave(file);

				}
				else
				{
					if (saveStandard)
					{
						gw.SaveStandard(file);
					}
					else
					{
						int maxdist  = 0;
						//here I need to use the input parameters if there are any...
						if (args.length > 1)
						{
							String nextArg  = args[1];
							maxdist = Integer.parseInt(nextArg);
						}

						buildWindow_(file, gw, maxdist);
					}
				}

			}
			System.exit(0);
		}
		else
		{

			Frame frame  = new Frame("TCS " + dna.VERSION);
			VGJ vgj      = new VGJ();
			vgj.isApp_ = true;
			vgj.init();
			vgj.start();

			//   vgj.buildWindow_();
			frame.add("Center", vgj);
			frame.resize(170, 100);
			frame.validate();
			frame.show();

		}
	}



	/**
	 *  Description of the Method
	 *
	 * @param  nextFile  Description of the Parameter
	 * @param  gw        Description of the Parameter
	 * @param  maxdist   Description of the Parameter
	 */
	private static void buildWindow_(String nextFile, GraphWindow gw, int maxdist)
	{
		// GraphWindow graph_editing_window  = new GraphWindow(true);
		gw.open(nextFile);
		if (maxdist > 0)
		{
			gw.SetMaxDistance(maxdist);
		}
		gw.RUN();

		gw.SaveStandard(nextFile + ".graph");
		//public void SetMaxDistance(int distance)

		gw.dispose();
		/*
		    GraphWindow graph_editing_window  = new GraphWindow(true);
		    graph_editing_window.open(nextFile);
		    graph_editing_window.RUN();
		    gw.SaveStandard(nextFile + ".graph");
		    //public void SetMaxDistance(int distance)
		    graph_editing_window.dispose();
		 */
	}


	/**
	 *  Description of the Method
	 *
	 * @param  bool  Description of the Parameter
	 * @return       Description of the Return Value
	 */
	private static GraphWindow buildWindow_(boolean bool)
	{

		// Bring up an undirected graph editor window.
		// The parameter to GraphWindow() indicates directed
		// or undirected.
		GraphWindow graph_editing_window  = new GraphWindow(true);

		// Here the algorithms are added.
		// graph_editing_window.addAlgorithmMenu("Tree");
		// ExampleAlg2 alg2 = new ExampleAlg2();
		// graph_editing_window.addAlgorithm(alg2, "Random");


		TreeAlgorithm talg                = new TreeAlgorithm('d');
		graph_editing_window.addAlgorithm(talg, "Tree Down");
		talg = new TreeAlgorithm('u');
		graph_editing_window.addAlgorithm(talg, "Tree Up");
		talg = new TreeAlgorithm('l');
		graph_editing_window.addAlgorithm(talg, "Tree Left");
		talg = new TreeAlgorithm('r');
		graph_editing_window.addAlgorithm(talg, "Tree Right");

		// if (appCount_++ == 0)
		// {
		// graph_editing_window.setTitle("TCS " + dna.VERSION);
		// }
		// else
		// {
		// graph_editing_window.setTitle("TCS " + dna.VERSION + ": " + appCount_);
		// }

		graph_editing_window.pack();
		graph_editing_window.show();

		return graph_editing_window;
	}



	//the version
	//public static String VERSION = "v1.10";
	/**
	 *  Description of the Method
	 *
	 * @param  event  Description of the Parameter
	 * @param  what   Description of the Parameter
	 * @return        Description of the Return Value
	 */
	public boolean action(Event event, Object what)
	{

		if (event.target instanceof Button)
		{

			if (((String) what).equals("Exit"))
			{
				System.exit(0);
			}
			else
			{
				buildWindow_();
			}
		}

		return super.action(event, what);
	}


	/**  Description of the Method */
	public void init()
	{

		if (isApp_)
		{
			setLayout(new GridLayout(2, 1));

			Button b  = new Button("Start New TCS Analysis");
			add(b);
			b.setFont(new Font("Charcoal", Font.PLAIN, 10));

			Button c  = new Button("Exit");
			add(c);
			c.setFont(new Font("Charcoal", Font.PLAIN, 10));
		}
		else
		{
			setLayout(new GridLayout(1, 1));

			Button b  = new Button("Start New TCS Analysis");
			add(b);
			b.setFont(new Font("Charcoal", Font.PLAIN, 10));
		}

		Node.setToolkit(getToolkit());

		if (!isApp_)
		{
			Node.setContext(getCodeBase());
			GraphWindow.setContext(getCodeBase());
		}
		validate();
		show();
	}



	/**
	 *  Description of the Method
	 *
	 * @return    Description of the Return Value
	 */
	private GraphWindow buildWindow_()
	{

		// Bring up an undirected graph editor window.
		// The parameter to GraphWindow() indicates directed
		// or undirected.
		GraphWindow graph_editing_window  = new GraphWindow(true);

		TreeAlgorithm talg                = new TreeAlgorithm('d');
		graph_editing_window.addAlgorithm(talg, "Tree Down");
		talg = new TreeAlgorithm('u');
		graph_editing_window.addAlgorithm(talg, "Tree Up");
		talg = new TreeAlgorithm('l');
		graph_editing_window.addAlgorithm(talg, "Tree Left");
		talg = new TreeAlgorithm('r');
		graph_editing_window.addAlgorithm(talg, "Tree Right");

		if (appCount_++ == 0)
		{
			graph_editing_window.setTitle("TCS " + dna.VERSION);
		}
		else
		{
			graph_editing_window.setTitle("TCS " + dna.VERSION + ": " + appCount_);
		}

		graph_editing_window.pack();
		graph_editing_window.show();

		return graph_editing_window;
	}
}
