package clad;

import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;
import EDU.auburn.VGJ.gui.GraphWindow;
import javax.swing.JOptionPane;

/**
 *  Description of the Class
 *
 * @author     wooo
 * @created    January 20, 2004
 */
public class FileReader
{
	private TextInputStream in;
	private GraphWindow frame;
	public int num;
	public int len;
	public int maxNameLen;
	private int[][] distance;
	private boolean lines;
	private boolean printseq;
	public Vector components;
	public Vector alltaxa;
	public Vector realtaxa;
	public String missingString;
	public boolean warnMissing;
	private Logger Log;


	/**
	 *  Constructor for the FileReader object
	 *
	 * @param  gw   Description of the Parameter
	 * @param  log  Description of the Parameter
	 */
	public FileReader(GraphWindow gw, Logger log)
	{
		frame = gw;
		Log = log;
		components = new Vector();
		alltaxa = new Vector();
		realtaxa = new Vector();

		maxNameLen = 0;
		lines = false;
		printseq = false;

		missingString = "";
		warnMissing = false;

	}


	/**
	 *  Description of the Method
	 *
	 * @param  infile  Description of the Parameter
	 */
	public void ReadDistanceFile(String infile)
	{

		//TaxaItem firsttaxa;
		in = new TextInputStream(infile);

		String line  = in.readLine().toUpperCase();

		if (line.startsWith("#NEXUS"))
		{
			ReadNexusDis();

			// Distances in NEXUS format, lower matrix
		}
		else
		{
			ReadPhylipDis(line);

			// Distances in PHYLIP format, lower matrix
		}

	}



	/**
	 *  Description of the Method
	 *
	 * @param  infile   Description of the Parameter
	 * @param  gapmode  Description of the Parameter
	 */
	public void ReadInputFile(String infile, boolean gapmode)
	{

		in = new TextInputStream(infile);

		// Check which format do we have, PHYLIP or NEXUS
		String line  = in.readLine().toUpperCase();

		if (line.startsWith("#NEXUS"))
		{
			ReadNexus(gapmode);

			// NEXUS format SEQUENTIAL
		}
		else
		{
			ReadPhylip(line, gapmode);

			// PHYLIP format SEQUENTIAL
		}

		in.close();

	}


	/**
	 *  Description of the Method
	 *
	 * @param  c  Description of the Parameter
	 */
	private void ReadMatrix(char c)
	{

		/*
		    Reads a matrix of distances in lower diagonal format
		  */
		//      Frame    frame = new Frame();
		int popA;
		int popB;
		String[] names     = new String[num];
		String nname       = "";
		distance = new int[num][num];
		maxNameLen = 0;
		Log.dprintln("\rI am reading this distance matrix:");

		for (popA = 0; popA < num; popA++)
		{
			names[popA] = in.readWord();

			//reads the seq names from the infile
			if (lines)
			{

				// we have stolen the first character of the fist name. Let's get it back
				nname = nname.valueOf(c);
				nname = nname.concat(names[0]);
				names[0] = nname;
				lines = false;
			}

			Log.dprint("\r" + names[popA] + "  ");

			if (names[popA] == null)
			{
				JOptionPane.showMessageDialog(frame, "There was an error reading the sequence names" + "\nin the pairwise distance matrix", "TCS warning", JOptionPane.WARNING_MESSAGE);

				//            frame.dispose();
				in.close();
				System.exit(0);
			}

			if (maxNameLen < names[popA].length())
			{
				maxNameLen = names[popA].length();
			}

			for (popB = 0; popB < popA; popB++)
			{
				distance[popA][popB] = distance[popB][popA] = in.readInt();

				if (distance[popA][popB] < 1)
				{
					JOptionPane.showMessageDialog(frame, "There was an error reading the distance matrix." + "\nAre you using absolute distances ?" + "\nAre you inputing HAPLOTYPE distances (no zeroes in the matrix)?", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}

				Log.dprint(" " + distance[popA][popB]);
			}
		}

		// Now convert the matrix to the vectors used in the rest of the program
		Log.dprintln();

		Component tempc;
		TaxaItem temptaxa;
		int cnt;

		for (cnt = 0; cnt < num; cnt++)
		{
			tempc = new Component(cnt);
			components.addElement(tempc);
			temptaxa = new TaxaItem(names[cnt], len, cnt);
			temptaxa.isIntermediate = false;
			tempc.taxa.addElement(temptaxa);
			temptaxa.parent = tempc;
			alltaxa.addElement(temptaxa);
			realtaxa.addElement(temptaxa);
		}

		Component tempc1;
		Component tempc2;
		TaxaItem tempt1;
		TaxaItem tempt2;
		Enumeration enum1  = components.elements();
		Distance mindist   = new Distance();

		while (enum1.hasMoreElements())
		{
			tempc1 = (Component) enum1.nextElement();
			tempt1 = (TaxaItem) tempc1.taxa.elementAt(0);
			tempc1.mindist.distance = dna.BIGNUM;

			Enumeration enum2  = components.elements();
			mindist.distance = dna.BIGNUM;

			// mindist will hold the minimum distance from c1 to any c2
			while (enum2.hasMoreElements())
			{
				tempc2 = (Component) enum2.nextElement();
				tempt2 = (TaxaItem) tempc2.taxa.elementAt(0);

				// tmpdist has the distance between these two
				Distance realdist    = new Distance(tempc1.id, tempc2.id, 0);

				// If it is myself, add the distance elements and go to top
				if (tempc1.id == tempc2.id)
				{

					Distance tmpcdist    = new Distance(tempc1.id, tempc2.id, 0, tempc1, tempc1);
					Distance tmptdist    = new Distance(tempt1.id, tempt2.id, 0, tempc1, tempc1);
					Distance metricdist  = new Distance(tempt1.id, tempt2.id, 0, tempc1, tempc1);
					tmpcdist.dc = tempc1;
					tmpcdist.sc = tempc1;
					tmptdist.dc = tempc1;
					tmptdist.sc = tempc1;
					realdist.dc = tempc1;
					realdist.sc = tempc1;
					tempc1.compdist.addElement(tmpcdist);
					tempt1.compdist.addElement(tmptdist);
					tempt1.realdist.addElement(realdist);
					tempt1.metricdist.addElement(metricdist);

					continue;
				}

				realdist.distance = distance[tempc1.id][tempc2.id];

				Distance tmpcdist    = new Distance(tempc1.id, tempc2.id, realdist.distance, tempc1, tempc2);
				Distance tmptdist    = new Distance(tempt1.id, tempt2.id, dna.BIGNUM, tempc1, tempc2);
				Distance metricdist  = new Distance(tempt1.id, tempt2.id, dna.BIGNUM, tempc1, tempc2);
				tempc1.compdist.addElement(tmpcdist);
				tempt1.compdist.addElement(tmptdist);
				tempt1.realdist.addElement(realdist);
				tempt1.metricdist.addElement(metricdist);

				if (realdist.distance < mindist.distance)
				{
					mindist.distance = realdist.distance;

					// mindist has taxa ids not component ids
					mindist.source = tempt1.id;
					mindist.sc = tempc1;
					mindist.destination = tempt2.id;
					mindist.dc = tempc2;
					Log.dprintln("min distance for " + tempc1.id + " = " + mindist.distance + " from " + mindist.source + " to " + mindist.destination, 0x20);
				}
			}

			Log.dprintln("final min distance for " + tempc1.id + " = " + mindist.distance + " from " + mindist.source + " to " + mindist.destination, 0x20);
			tempc1.mindist.clone(mindist);
			tempt1.minRealDist = mindist.distance;
		}
	}


	/**
	 *  Description of the Method
	 *
	 * @param  gapmode  Description of the Parameter
	 */
	private void ReadNexus(boolean gapmode)
	{

		/*
		    Reads NEXUS format SEQUENTIAL
		  */
		char c;
		int j;
		int cnt                 = 0;
		boolean missing         = false;
		String name             = "";
		Log.dprintln("Data in NEXUS format");

		String token            = in.readWord().toUpperCase();

		while (!token.startsWith("NTAX="))
		{
			token = in.readWord().toUpperCase();
		}

		StringTokenizer reader  = new StringTokenizer(token, " =\n\r\t;");
		reader.nextToken();
		num = Integer.parseInt(reader.nextToken());

		while (!token.startsWith("NCHAR="))
		{
			token = in.readWord().toUpperCase();
		}

		reader = new StringTokenizer(token, " =\n\r\t;");
		reader.nextToken();
		len = Integer.parseInt(reader.nextToken());
		Log.dprintln("Number of sequences: " + num);
		Log.dprintln("Length of sequences: " + len);

		while (!token.equals("MATRIX"))
		{

			// go all the way to the sequences
			token = in.readWord().toUpperCase();
		}

		// get rid of potential stuff after the "Matrix" word and before sequence name
		c = in.readChar();

		while (c == '\r' || c == '\n' || c == '\t')
		{
			c = in.readChar();
			lines = true;
		}

		// if there is a comment for the position numbers (e.g McClade) [ 1 2 3 4] after "Matrix"
		if (c == '[')
		{

			while ((c = in.readChar()) != ']')
			{
			}
		}

		// in case there are more things in the way to the sequences
		while (c == '\r' || c == '\n' || c == '\t' || c == ']' || c == ' ')
		{
			c = in.readChar();
			lines = true;
			Log.dprintln(c);
		}

		for (int i = 0; i < num; i++)
		{
			name = in.readWord();

			if (lines)
			{

				// we have stolen the first character of the fist name. Let's get it back
				String nname  = String.valueOf(c);
				nname = nname.concat(name);
				name = nname;
				lines = false;
			}

			if (printseq)
			{
				Log.dprintln("\r" + (i + 1) + ". " + name);
			}

			TaxaItem temptaxa   = new TaxaItem(name, len, cnt);
			temptaxa.isIntermediate = false;

			if (name.length() == 0)
			{
				name = in.readLine();
			}

			if (maxNameLen < name.length())
			{
				maxNameLen = name.length();
			}

			temptaxa.name = name;

			for (j = 0; j < len; j++)
			{

				// skip white spaces between name and beginning of sequence
				while ((c = in.readChar()) == ' ')
				{
				}

				c = Character.toUpperCase(c);

				/*
				    DP 240201
				  */
				temptaxa.characters[j] = c;

				if (printseq)
				{
					System.err.print(c);
				}

				if (c == '\n' || c == '\r')
				{
					JOptionPane.showMessageDialog(frame, "You have extra returns in sequence \"" + name + "\".", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}

				if (c == ' ')
				{
					JOptionPane.showMessageDialog(frame, "The lenght of sequence \"" + name + "\"" + " is not " + len + ".", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}

				if (c != 'A' && c != 'C' && c != 'G' && c != 'T' && c != '-' && c != '.' && c != 'a' && c != 'c' && c != 'g' && c != 't' && c != 'N' && c != '?')
				{
					JOptionPane.showMessageDialog(frame, "One of the characters in sequence \"" + name + "\" is not (ACGCacgt.?N-).", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}

				if (in.EOF())
				{
					JOptionPane.showMessageDialog(frame, "Unexpected end of file was encountered.", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}
			}

			Log.dprintln("Read data: " + temptaxa.characters, 1);
			in.readLine();

			// get rid of the CR
			TaxaItem firsttaxa;
			int numdups         = 0;
			char newc;
			char oldc;
			Enumeration enumc   = components.elements();
			Component tempc;

			while (enumc.hasMoreElements())
			{
				firsttaxa = (TaxaItem) ((Component) components.elementAt(0)).taxa.elementAt(0);
				tempc = (Component) enumc.nextElement();
				Log.dprintln("looking at component: " + tempc.id, 1);

				boolean dup     = true;
				missing = false;
				TaxaItem tempt;
				tempt = (TaxaItem) tempc.taxa.elementAt(0);

				// There should be exactly one here
				Log.dprintln("looking at taxa: " + tempt.id + " named " + tempt.name + "and" + temptaxa.id + " named " + temptaxa.name, 1);
				Log.dprintln("comparing " + tempt.characters, 1);
				Log.dprintln("with " + temptaxa.characters, 1);

				//compare the all elements of components TaxaItem w/ the new taxaItem
				for (int k = 0; k < len; k++)
				{
					newc = temptaxa.characters[k];
					oldc = tempt.characters[k];

					if (newc == '.')
					{
						newc = firsttaxa.characters[k];
					}

					if (oldc == '.')
					{
						oldc = firsttaxa.characters[k];
					}

					Log.dprint("[" + newc + "," + oldc + "]", 1);

					if (!gapmode && (newc == '-' || oldc == '-'))
					{

						continue;
					}

					if (newc == '?' || oldc == '?' || newc == 'N' || oldc == 'N')
					{
						missing = true;

						continue;
					}

					if (newc != oldc)
					{
						dup = false;

						break;
					}
				}

				if (dup && missing)
				{
					warnMissing = true;
					missingString = missingString.concat(tempt.name + " and " + temptaxa.name + " differ only by missing or ambiguous characters\n"/* + new String(tempt.characters) + " vs " + new String(temptaxa.characters)*/);
					// print the full seqs...
//					System.out.println("tempt");
				}

				Log.dprintln(1);

				if (dup == true)
				{
					tempt.numduplicates++;
					numdups++;
					tempt.dupnames.addElement(temptaxa.name);
					Log.dprintln("duplicate name: " + name + " duplicates " + tempt.numduplicates, 1);

					break;
				}
			}

			if (numdups == 0)
			{

				//this is a new taxa not equal to any other
				// Add this to the component Vector
				tempc = new Component(cnt);
				tempc.taxa.addElement(temptaxa);
				temptaxa.parent = tempc;
				alltaxa.addElement(temptaxa);
				realtaxa.addElement(temptaxa);
				cnt++;
				components.addElement(tempc);
				Log.dprintln("Added name: " + name, 1);
			}
		}
	}


	/**  Description of the Method */
	private void ReadNexusDis()
	{

		char c;

		//      Frame frame = new Frame();
		Log.dprintln("\rReading NEXUS distance file\r");

		//this function parses out the header of the nex files
		String token            = in.readWord().toUpperCase();

		while (!token.startsWith("NTAX="))
		{
			token = in.readWord().toUpperCase();
		}

		StringTokenizer reader  = new StringTokenizer(token, " =\n\r\t;");
		reader.nextToken();
		num = Integer.parseInt(reader.nextToken());

		while (!token.startsWith("NCHAR="))
		{
			token = in.readWord().toUpperCase();
		}

		reader = new StringTokenizer(token, " =\n\r\t;");
		reader.nextToken();
		len = Integer.parseInt(reader.nextToken());

		if (Character.isDigit((char) num))
		{
			JOptionPane.showMessageDialog(frame, "Problems reading the number of the sequences!", "TCS warning", JOptionPane.WARNING_MESSAGE);

			//         frame.dispose();
			in.close();
			System.exit(0);
		}

		if (Character.isDigit((char) len))
		{
			JOptionPane.showMessageDialog(frame, "Problems reading the length of the sequences!", "TCS warning", JOptionPane.WARNING_MESSAGE);

			//         frame.dispose();
			in.close();
			System.exit(0);
		}

		Log.dprintln("Number of sequences: " + num);
		Log.dprintln("Length of sequences: " + len);

		while (!token.equals("MATRIX"))
		{

			// go all the way to the distances
			token = in.readWord().toUpperCase();
		}

		// get rid of potential stuff after the "Matrix" word and before sequence name
		c = in.readChar();

		while (c == '\r' || c == '\n' || c == '\t')
		{
			c = in.readChar();
			lines = true;
		}

		// if there is a comment for the position numbers (e.g McClade) [ 1 2 3 4] after "Matrix"
		if (c == '[')
		{

			while ((c = in.readChar()) != ']')
			{
			}
		}

		// in case there are more things in the way to the sequences
		while (c == '\r' || c == '\n' || c == '\t' || c == ']' || c == ' ')
		{
			c = in.readChar();
			lines = true;
			Log.dprintln(c);
		}

		ReadMatrix(c);
		in.close();
	}

	//FIXME:  Put this in another class!
	/**
	 *  Description of the Method
	 *
	 * @param  line     Description of the Parameter
	 * @param  gapmode  Description of the Parameter
	 */
	private void ReadPhylip(String line, boolean gapmode)
	{

		/*
		    Reads PHYLIP format SEQUENTIAL (max length of the names = 10)
		  */
		char c;
		char ch;
		int j;
		TaxaItem firsttaxa;
		boolean missing         = false;
		int cnt                 = 0;
		boolean cr              = false;
		Log.dprintln("\nData in PHYLIP format");

		StringTokenizer reader  = new StringTokenizer(line);
		num = Integer.parseInt(reader.nextToken());
		len = Integer.parseInt(reader.nextToken());
		Log.dprintln("Number of sequences: " + num);
		Log.dprintln("Length of sequences: " + len);

		// Add the components
		for (int i = 0; i < num; i++)
		{

			StringBuffer label  = new StringBuffer();

			for (int k = 0; k < 10; k++)
			{
				ch = in.readChar();

				if (ch == '\n' || ch == '\r')
				{
					cr = true;

					break;
				}

				label.append(ch);
			}

			String name         = label.toString();

			if (printseq)
			{
				Log.dprintln("\r" + (i + 1) + ". " + name);
			}

			TaxaItem temptaxa   = new TaxaItem(name, len, cnt);
			temptaxa.isIntermediate = false;

			if (name.length() == 0)
			{
				name = in.readLine();
			}

			if (maxNameLen < name.length())
			{
				maxNameLen = name.length();
			}

			temptaxa.name = name;

			for (j = 0; j < len; j++)
			{
				c = in.readChar();

				if (c == '\n' || c == '\r')
				{

					// there is a CR after the name
					c = in.readChar();
				}

				c = Character.toUpperCase(c);

				/*
				    DP 240201
				  */
				temptaxa.characters[j] = c;

				if (printseq)
				{
					System.err.print(c);
				}

				if (c == '\n' || c == '\r')
				{
					JOptionPane.showMessageDialog(frame, "You have extra returns in sequence \"" + name + "\".", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}

				if (c == ' ')
				{
					JOptionPane.showMessageDialog(frame, "The lenght of sequence \"" + name + "\"" + " is not " + len + ".", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}

				if (c != 'A' && c != 'C' && c != 'G' && c != 'T' && c != '-' && c != '.' && c != 'a' && c != 'c' && c != 'g' && c != 't' && c != 'N' && c != '?')
				{
					JOptionPane.showMessageDialog(frame, "One of the characters in sequence \"" + name + "\" is not (ACGCacgt.?N-).", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}

				if (in.EOF())
				{
					JOptionPane.showMessageDialog(frame, "Unexpected end of file was encountered.", "TCS warning", JOptionPane.WARNING_MESSAGE);

					//               frame.dispose();
					in.close();
					System.exit(0);
				}
			}

			Log.dprintln("Read data: " + temptaxa.characters, 1);
			in.readLine();

			// get rid of the CR
			int numdups         = 0;
			char newc;
			char oldc;
			Enumeration enumc   = components.elements();
			Component tempc;

			while (enumc.hasMoreElements())
			{
				firsttaxa = (TaxaItem) ((Component) components.elementAt(0)).taxa.elementAt(0);
				tempc = (Component) enumc.nextElement();
				Log.dprintln("looking at component: " + tempc.id, 1);

				boolean dup     = true;
				missing = false;
				TaxaItem tempt;
				tempt = (TaxaItem) tempc.taxa.elementAt(0);

				// There should be exactly one here
				Log.dprintln("looking at taxa: " + tempt.id + " named " + tempt.name + "and" + temptaxa.id + " named " + temptaxa.name, 1);
				Log.dprintln("comparing " + tempt.characters, 1);
				Log.dprintln("with " + temptaxa.characters, 1);

				for (int k = 0; k < len; k++)
				{
					newc = temptaxa.characters[k];
					oldc = tempt.characters[k];

					if (newc == '.')
					{
						newc = firsttaxa.characters[k];
					}

					if (oldc == '.')
					{
						oldc = firsttaxa.characters[k];
					}

					Log.dprint("[" + newc + "," + oldc + "]", 1);

					if (!gapmode && (newc == '-' || oldc == '-'))
					{

						continue;
					}

					if (newc == '?' || oldc == '?' || newc == 'N' || oldc == 'N')
					{
						missing = true;

						continue;
					}

					if (newc != oldc)
					{
						dup = false;

						break;
					}
				}

				if (dup && missing)
				{
					warnMissing = true;
					missingString = missingString.concat(tempt.name + " and " + temptaxa.name + " differ only by missing or ambiguous characters\r");
				}

				Log.dprintln(1);

				if (dup == true)
				{
					tempt.numduplicates++;
					numdups++;
					tempt.dupnames.addElement(temptaxa.name);
					Log.dprintln("duplicate name: " + name + " duplicates " + tempt.numduplicates, 1);

					break;
				}
			}

			if (numdups == 0)
			{

				// Add this to the component Vector
				tempc = new Component(cnt);
				tempc.taxa.addElement(temptaxa);
				temptaxa.parent = tempc;
				alltaxa.addElement(temptaxa);
				realtaxa.addElement(temptaxa);
				cnt++;
				components.addElement(tempc);
				Log.dprintln("Added name: " + name, 1);
			}
		}
	}


	/**
	 *  Description of the Method
	 *
	 * @param  line  Description of the Parameter
	 */
	private void ReadPhylipDis(String line)
	{
		Log.dprintln("\rReading PHYLIP distance file\r");

		StringTokenizer reader  = new StringTokenizer(line);

		//      Frame           frame = new Frame();
		num = Integer.parseInt(reader.nextToken());
		len = Integer.parseInt(reader.nextToken());

		if (Character.isDigit((char) num))
		{
			JOptionPane.showMessageDialog(frame, "Problems reading the number of the sequences!", "TCS warning", JOptionPane.WARNING_MESSAGE);

			//         frame.dispose();
			in.close();
			System.exit(0);
		}

		if (Character.isDigit((char) len))
		{
			JOptionPane.showMessageDialog(frame, "Problems reading the length of the sequences!", "TCS warning", JOptionPane.WARNING_MESSAGE);

			//         frame.dispose();
			in.close();
			System.exit(0);
		}

		Log.dprintln("Number of sequences: " + num);
		Log.dprintln("Length of sequences: " + len);
		ReadMatrix('a');
		in.close();
	}

}
