package byu;

/****************************************************************************
** COPYRIGHT (C):    1996 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Computing Concepts with Java
** FILE:             Circle.java
****************************************************************************/

/**
 * A circle shape for simple graphics
 * @version 1.00 11 Apr 1997
 * @author Cay Horstmann
 * @author Quinn Snell (BYU)
 */

public class Circle extends Shape
{  /**
	* Constructs a circle
	* @param c the center of the circle
	* @param r the radius of the circle
	*/

   Point center;
   double radius;
   public Circle(Point c, double r)
   {  center = c;
	  radius = r;
   }   
   /**
	* Creates a copy of this object
	* @return a copy of this object
	*/

   public Object clone()
   {  Circle r = (Circle)super.clone();
	  r.center = (Point)center.clone();
	  return r;
   }   
   public void draw()
   {
	  GraphicsApplet.addShape(this);
   }   
   /**
	* Tests if this object equals another
	* @param b the object to compare with
	* @return <tt>true</tt> if the two objects are equal
	*/

   public boolean equals(Object b)
   {  if (b instanceof Circle)
	  {  Circle pb = (Circle)b;
		 return center.equals(pb.center) && radius == pb.radius;
	  }
	  else return false;
   }   
   public void erase()
   {
	  GraphicsApplet.erase(this);
   }   
   /**
	* Gets the center of the circle
	* @return the center of the circle
	*/

   public Point getCenter()
   {  return center;
   }   
   /**
	* Gets the radius of the circle
	* @return the radius of the circle
	*/

   public double getRadius()
   {  return radius;
   }   
   /**
	* Computes a hash code for this object
	* @return a hash code for this object
	*/

   public int hashCode()
   {  return (int)(19 * center.hashCode() + 23 * radius);
   }   
   /**
	* Moves this graphical shape
	* @param dx the amount to move by in x direction
	* @param dy the amount to move by in y direction
	*/

   public void move(double x, double y)
   {  center.move(x, y);
   }   
   public void redraw(java.awt.Graphics g, GraphicsCoord coord)
   {  int xrad = Math.abs(coord.xToPixelDist(radius));
	  int yrad = Math.abs(coord.yToPixelDist(radius));
	  int xtop = coord.xToPixelPos(center.getX()) - xrad;
	  int ytop = coord.yToPixelPos(center.getY()) - yrad;
	  g.drawOval(xtop, ytop, 2 * xrad, 2 * yrad);
   }   
   /**
	* Creates a string representation of this object
	* @return a string representation of this object
	*/

   public String toString()
   {  return "Circle[center = " + center.toString() + "," + "radius = " + radius + "]" ;
   }   
}
