package byu;

/****************************************************************************
** COPYRIGHT (C):    1996 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Computing Concepts with Java
** FILE:             Line.java
****************************************************************************/

/**
 * A line shape for simple graphics
 * @version 1.00 11 Apr 1997
 * @author Cay Horstmann
 * @author Quinn Snell (BYU)
 */

public class Line extends Shape
{  /**
	* Constructs a line object
	* @param s the starting point of the line
	* @param e the ending point of the line
	*/

   private Point start;
   private Point end;
   public Line(Point s, Point e)
   {  start = s;
	  end = e;
   }   
   /**
	* Creates a copy of this object
	* @return a copy of this object
	*/

   public Object clone()
   {  Line r = (Line)super.clone();
	  r.start = (Point)start.clone();
	  r.end = (Point)end.clone();
	  return r;
   }   
   public void draw()
   {
	  GraphicsApplet.addShape(this);
   }   
   /**
	* Tests if this object equals another
	* @param b the object to compare with
	* @return <tt>true</tt> if the two objects are equal
	*/

   public boolean equals(Object b)
   {  if (b instanceof Line)
	  {   Line pb = (Line)b;
		  return start.equals(pb.start) && end.equals(pb.end);
	  }
	  else return false;
   }   
   public void erase()
   {
	  GraphicsApplet.erase(this);
   }   
   /**
	* Gets the ending point of the line
	* @return the ending point of the line
	*/

   public Point getEnd()
   {  return end;
   }   
   /**
	* Gets the starting point of the line
	* @return the starting point of the line
	*/

   public Point getStart()
   {  return start;
   }   
   /**
	* Computes a has code for this object
	* @return a hash code for this object
	*/

   public int hashCode()
   {  return 13 * start.hashCode() + 17 * end.hashCode();
   }   
   /**
	* Moves this graphical shape
	* @param dx the amount to move by in x direction
	* @param dy the amount to move by in y direction
	*/

   public void move(double x, double y)
   {  start.move(x, y);
	  end.move(x, y);
   }   
   void redraw(java.awt.Graphics g, GraphicsCoord coord)
   {  g.drawLine(coord.xToPixelPos(start.getX()), coord.yToPixelPos(start.getY()), 
		 coord.xToPixelPos(end.getX()), coord.yToPixelPos(end.getY()));
   }   
   /**
	* Creates a string representation of this object
	* @return a string representation of this object
	*/

   public String toString()
   {  return "Line[start = " + start.toString() + "," + "end = " + end.toString() + "]";
   }   
}
